@@ -2,7 +2,9 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this file,
 # You can obtain one at http://mozilla.org/MPL/2.0/.
 
+import os
 import json
+import jwt
 import responses
 import six
 try:
@@ -22,6 +24,13 @@ from six.moves.urllib.parse import urlparse, parse_qs
 TEST_SERVER_URL = "https://server/v1"
 
 
+def add_jwks_response():
+    responses.add(responses.GET,
+                  'https://server/v1/jwks',
+                  body=open(os.path.join(os.path.dirname(__file__), "jwks.json")).read(),
+                  content_type='application/json')
+
+
 class TestClientServerUrl(unittest.TestCase):
     def test_trailing_slash_without_prefix_added_prefix(self):
         client = Client('abc', 'cake', "https://server/")
@@ -165,9 +174,9 @@ class TestAuthClientVerifyCode(unittest.TestCase):
                       'https://server/v1/verify',
                       body=body,
                       content_type='application/json')
-
+        add_jwks_response()
         self.verification = self.client.verify_token(token='abc')
-        self.response = responses.calls[0]
+        self.response = responses.calls[1]
 
     def test_reaches_server_on_verify_url(self):
         self.assertEqual(self.response.request.url,
@@ -194,6 +203,7 @@ class TestAuthClientVerifyCode(unittest.TestCase):
                       'https://server/v1/verify',
                       body='{"missing": "attributes"}',
                       content_type='application/json')
+        add_jwks_response()
         self.assertRaises(fxa.errors.OutOfProtocolError,
                           self.client.verify_token,
                           token='1234')
@@ -205,6 +215,7 @@ class TestAuthClientVerifyCode(unittest.TestCase):
                       'https://server/v1/verify',
                       body=body,
                       content_type='application/json')
+        add_jwks_response()
         self.assertRaises(fxa.errors.ScopeMismatchError,
                           self.client.verify_token,
                           token='1234',
@@ -272,35 +283,44 @@ class TestAuthClientAuthorizeCode(unittest.TestCase):
     server_url = TEST_SERVER_URL
 
     def setUp(self):
+        def authorization_callback(request):
+            data = json.loads(_decoded(request.body))
+            headers = {
+                'Content-Type': 'application/json'
+            }
+            body = {
+                'redirect': 'https://relier/page?code=qed&state={}'.format(data["state"])
+            }
+            return (200, headers, json.dumps(body))
+
         self.client = Client("abc", "xyz", server_url=self.server_url)
-        body = '{"redirect": "https://relier/page?code=qed&state=blah"}'
-        responses.add(responses.POST,
-                      'https://server/v1/authorization',
-                      body=body,
-                      content_type='application/json')
+        responses.add_callback(responses.POST,
+                               'https://server/v1/authorization',
+                               callback=authorization_callback,
+                               content_type='application/json')
 
     @responses.activate
     def test_authorize_code_with_default_arguments(self):
         assertion = "A_FAKE_ASSERTION"
         code = self.client.authorize_code(assertion)
-        self.assertEquals(code, "qed")
+        self.assertEqual(code, "qed")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": assertion,
             "client_id": self.client.client_id,
-            "state": "x",
+            "state": AnyStringValue(),
         })
 
     @responses.activate
     def test_authorize_code_with_explicit_scope(self):
         assertion = "A_FAKE_ASSERTION"
         code = self.client.authorize_code(assertion, scope="profile:email")
-        self.assertEquals(code, "qed")
+        self.assertEqual(code, "qed")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": assertion,
             "client_id": self.client.client_id,
-            "state": "x",
+            "state": AnyStringValue(),
             "scope": "profile:email",
         })
 
@@ -308,12 +328,12 @@ class TestAuthClientAuthorizeCode(unittest.TestCase):
     def test_authorize_code_with_explicit_client_id(self):
         assertion = "A_FAKE_ASSERTION"
         code = self.client.authorize_code(assertion, client_id="cba")
-        self.assertEquals(code, "qed")
+        self.assertEqual(code, "qed")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": assertion,
             "client_id": "cba",
-            "state": "x",
+            "state": AnyStringValue(),
         })
 
     @responses.activate
@@ -325,12 +345,12 @@ class TestAuthClientAuthorizeCode(unittest.TestCase):
         self.assertEqual(sorted(verifier),
                          ["code_verifier"])
         code = self.client.authorize_code(assertion, **challenge)
-        self.assertEquals(code, "qed")
+        self.assertEqual(code, "qed")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": assertion,
             "client_id": self.client.client_id,
-            "state": "x",
+            "state": AnyStringValue(),
             "code_challenge": challenge["code_challenge"],
             "code_challenge_method": challenge["code_challenge_method"],
         })
@@ -344,12 +364,12 @@ class TestAuthClientAuthorizeCode(unittest.TestCase):
             audience=TEST_SERVER_URL,
             service=self.client.client_id
         )
-        self.assertEquals(code, "qed")
+        self.assertEqual(code, "qed")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": "IDENTITY",
             "client_id": self.client.client_id,
-            "state": "x",
+            "state": AnyStringValue(),
         })
 
 
@@ -363,17 +383,18 @@ class TestAuthClientAuthorizeToken(unittest.TestCase):
                       'https://server/v1/authorization',
                       body='{"access_token": "izatoken"}',
                       content_type='application/json')
+        add_jwks_response()
 
     @responses.activate
     def test_authorize_token_with_default_arguments(self):
         assertion = "A_FAKE_ASSERTION"
         token = self.client.authorize_token(assertion)
-        self.assertEquals(token, "izatoken")
+        self.assertEqual(token, "izatoken")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": assertion,
             "client_id": self.client.client_id,
-            "state": "x",
+            "state": AnyStringValue(),
             "response_type": "token",
         })
 
@@ -381,12 +402,12 @@ class TestAuthClientAuthorizeToken(unittest.TestCase):
     def test_authorize_token_with_explicit_scope(self):
         assertion = "A_FAKE_ASSERTION"
         token = self.client.authorize_token(assertion, scope="storage")
-        self.assertEquals(token, "izatoken")
+        self.assertEqual(token, "izatoken")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": assertion,
             "client_id": self.client.client_id,
-            "state": "x",
+            "state": AnyStringValue(),
             "response_type": "token",
             "scope": "storage",
         })
@@ -395,12 +416,12 @@ class TestAuthClientAuthorizeToken(unittest.TestCase):
     def test_authorize_token_with_explicit_client_id(self):
         assertion = "A_FAKE_ASSERTION"
         token = self.client.authorize_token(assertion, client_id="cba")
-        self.assertEquals(token, "izatoken")
+        self.assertEqual(token, "izatoken")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": assertion,
             "client_id": "cba",
-            "state": "x",
+            "state": AnyStringValue(),
             "response_type": "token",
         })
 
@@ -413,12 +434,12 @@ class TestAuthClientAuthorizeToken(unittest.TestCase):
             audience=TEST_SERVER_URL,
             service=self.client.client_id
         )
-        self.assertEquals(token, "izatoken")
+        self.assertEqual(token, "izatoken")
         req_body = json.loads(_decoded(responses.calls[0].request.body))
-        self.assertEquals(req_body, {
+        self.assertEqual(req_body, {
             "assertion": "IDENTITY",
             "client_id": self.client.client_id,
-            "state": "x",
+            "state": AnyStringValue(),
             "response_type": "token",
         })
 
@@ -530,6 +551,7 @@ class TestCachedClient(unittest.TestCase):
                       'https://server/v1/verify',
                       body=self.body,
                       content_type='application/json')
+        add_jwks_response()
 
     def test_has_default_cache(self):
         self.assertIsNotNone(self.client.cache)
@@ -591,3 +613,136 @@ class TestGeventPatch(unittest.TestCase):
         self.assertEqual(fxa._utils.requests, grequests)
 
         fxa._utils.requests = old_requests
+
+
+class TestJwtToken(unittest.TestCase):
+
+    server_url = TEST_SERVER_URL
+
+    def callback(self, request):
+        if self.verify_will_succeed:
+            return (200, {}, self.body)
+        return (500, {}, '{}')
+
+    def _make_jwt(self, payload, key, alg="RS256", header={"typ": "at+jwt"}):
+        header = header.copy()  # So we don't accidentally mutate argument in-place
+        return six.ensure_text(jwt.encode(payload, key, alg, header))
+
+    def setUp(self):
+        self.client = Client(server_url=self.server_url)
+        self.body = ('{"user": "alice", "scope": ["profile"],'
+                     '"client_id": "abc"}')
+        responses.add_callback(responses.POST,
+                               'https://server/v1/verify',
+                               callback=self.callback,
+                               content_type='application/json')
+        add_jwks_response()
+        self.verify_will_succeed = True
+
+    def get_file_contents(self, filename):
+        return jwt.algorithms.RSAAlgorithm.from_jwk(open(
+            os.path.join(
+                os.path.dirname(__file__),
+                filename
+            )
+        ).read())
+
+    @responses.activate
+    def test_good_jwt_token(self):
+        private_key = self.get_file_contents("private-key.json")
+        token = self._make_jwt({
+            "sub": "asdf",
+            "scope": "qwer",
+            "client_id": "foo"
+        }, private_key)
+        self.client.verify_token(token)
+        for c in responses.calls:
+            if c.request.url == 'https://server/v1/verify':
+                raise Exception("testing with a good token should not have \
+                                 resulted in a call to /verify, but it did.")
+
+    @responses.activate
+    def test_good_jwt_token_with_long_form_typ_header(self):
+        private_key = self.get_file_contents("private-key.json")
+        token = self._make_jwt({
+            "sub": "asdf",
+            "scope": "qwer",
+            "client_id": "foo"
+        }, private_key, header={
+            "typ": "application/at+JWT",
+        })
+        self.client.verify_token(token)
+        for c in responses.calls:
+            if c.request.url == 'https://server/v1/verify':
+                raise Exception("testing with a good token should not have \
+                                 resulted in a call to /verify, but it did.")
+
+    @responses.activate
+    def test_good_jwt_token_with_correct_scope(self):
+        private_key = self.get_file_contents("private-key.json")
+        token = self._make_jwt({
+            "sub": "asdf",
+            "scope": "qwer tee",
+            "client_id": "foo"
+        }, private_key)
+        self.client.verify_token(token, scope="tee")
+
+    @responses.activate
+    def test_good_jwt_token_with_incorrect_scope(self):
+        private_key = self.get_file_contents("private-key.json")
+        token = self._make_jwt({
+            "sub": "asdf",
+            "scope": "qwer",
+            "client_id": "foo"
+        }, private_key)
+        with self.assertRaises(fxa.errors.TrustError):
+            self.client.verify_token(token, scope="tee")
+
+    @responses.activate
+    def test_wrong_key_jwt_token(self):
+        self.verify_will_succeed = False
+        bad_key = self.get_file_contents("bad-key.json")
+        token = self._make_jwt({}, bad_key)
+        with self.assertRaises(fxa.errors.TrustError):
+            self.client.verify_token(token)
+        for c in responses.calls:
+            if c.request.url == 'https://server/v1/verify':
+                raise Exception("testing with a well-formed token with invalid signature \
+                                 should not have resulted in a call to /verify, but it did.")
+
+    @responses.activate
+    def test_expired_jwt_token(self):
+        private_key = self.get_file_contents("private-key.json")
+        token = self._make_jwt({"qwer": "asdf", "exp": 0}, private_key)
+        with self.assertRaises(fxa.errors.TrustError):
+            self.client.verify_token(token)
+
+    @responses.activate
+    def test_jwt_token_with_wrong_typ(self):
+        private_key = self.get_file_contents("private-key.json")
+        token = self._make_jwt({"qwer": "asdf", "exp": 0}, private_key, header={
+          "typ": "rt+jwt"
+        })
+        with self.assertRaises(fxa.errors.TrustError):
+            self.client.verify_token(token)
+
+    @responses.activate
+    def test_garbage_jwt_token(self):
+        self.verify_will_succeed = False
+        with self.assertRaises(fxa.errors.ServerError):
+            self.client.verify_token("garbage")
+        for c in responses.calls:
+            if c.request.url == 'https://server/v1/verify':
+                break
+        else:
+            raise Exception("testing with a garbage token should have \
+                             called /verify, but it did not.")
+
+
+class AnyStringValue:
+
+    def __eq__(self, other):
+        return isinstance(other, six.string_types)
+
+    def __repr__(self):
+        return 'any string'
